'use strict';

/**
 * @class ModulusChecker
 *
 */
const _ = require('lodash'),
    ModulusWeights = require('../data/ModulusWeights'),
    SortCodeSubstitutions = require('../data/SortCodeSubstitutions');
class ModulusChecker {

    /**
     * Validate concatenated sort code and bank account number
     * @param  {String}   sortCode             The sort code
     * @param  {String}   bankAccountNumber    The bank account number
     *
     * @return  {Boolean}  True if check passed
     */
    check(sortCode, bankAccountNumber) {
        const sortCodeInt = parseInt(sortCode, 10),
            checkRecords = _.filter(ModulusWeights.data, function(moduleWeight) {
                return parseInt(moduleWeight.start, 10) <= sortCodeInt && parseInt(moduleWeight.end, 10) >= sortCodeInt;
            }),
            modulusCheckSubstitution = _.find(SortCodeSubstitutions.data, function(sortCodeSubstitution) {
                return sortCodeSubstitution.original === sortCode;
            });

        if (_.isEmpty(checkRecords) || checkRecords.length > 2) {
            return true;
        }
        const identificationNumberInNotation = getFormattedIdentificationNumber(sortCode, bankAccountNumber);
        let valid = true;

        if (checkRecords[0].exception === 14 && checkRecords.length === 1) {
            checkRecords.push(checkRecords[0]);
        }

        _.each(checkRecords, function(checkRecord, index) {
            const previousCheckRecord = index > 0 ? checkRecords[index - 1] : null;
            valid = validateByCheckRecord(identificationNumberInNotation, checkRecord, modulusCheckSubstitution, previousCheckRecord, valid);
        });

        return valid;
    }
}

module.exports = ModulusChecker;

/**
 * Get the formatted identification number to validate
 * @param  {String}   sortCode             The sort code
 * @param  {String}   bankAccountNumber    The bank account number
 *
 * @return  {Object}  the formatted identification number in notation form
 */
function getFormattedIdentificationNumber(sortCode, bankAccountNumber) {
    bankAccountNumber = replaceAll(bankAccountNumber, '-', '');
    sortCode = replaceAll(sortCode, '-', '');
    const bankAccountNumberLength = bankAccountNumber.length;

    if (bankAccountNumberLength > 8) {
        bankAccountNumber = bankAccountNumber.slice(9);
    }
    if (bankAccountNumberLength < 8) {
        bankAccountNumber = _.padLeft(bankAccountNumber, 8, '0');
    }
    const identificationNumber = sortCode + '' + bankAccountNumber;

    return convertToNotation(identificationNumber);
}

/**
 * Validate given id number in notation by given check number
 * @param  {Object}   identificationNumberInNotation   The identification number in modulus checking notation
 * @param  {Object}   checkRecord                      The weight record
 * @param  {Object}   modulusCheckSubstitution         The current check substitution record
 * @param  {Number}   modulusBy                        The number to modulus by
 * @param  {Object}   [previousCheckRecord]            The previous check record (if relevant)
 * @param  {Boolean}  [valid]                          True if valid in previous checks
 *
 * @return  {Boolean}  True if valid
 */
function validateByCheckRecord(identificationNumberInNotation, checkRecord, modulusCheckSubstitution, previousCheckRecord, valid) {
    let modulusCheckPassed;

    switch (checkRecord.mod_check) {
        case 'MOD10':
        case 'DBLAL':
            modulusCheckPassed = modulusCheck(identificationNumberInNotation, checkRecord, modulusCheckSubstitution, 10,
                previousCheckRecord, valid);
            break;
        case 'MOD11':
            modulusCheckPassed = modulusCheck(identificationNumberInNotation, checkRecord, modulusCheckSubstitution, 11,
                previousCheckRecord, valid);
            break;
    }

    if (!previousCheckRecord) {
        return valid && modulusCheckPassed;
    } else if ((previousCheckRecord.exception === 10 && checkRecord.exception === 11) ||
        (previousCheckRecord.exception === 12 && checkRecord.exception === 13) ||
        (previousCheckRecord.exception === 14)) {
        return valid || modulusCheckPassed;
    } else {
        return valid && modulusCheckPassed;
    }
}

/**
 * Convert the given identification number to the modulus checking validation
 * @param  {String}   identificationNumber             The identification number
 *
 * @return  {Object}  The identification number in modulus checking notation
 */
function convertToNotation(identificationNumber) {
    const identificationNumberArray = parseNumberArray(_.toArray(identificationNumber)),
        identificationNumberInNotation = {},
        notationCharacters = ['u', 'v', 'w', 'x', 'y', 'z', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'];

    _.each(identificationNumberArray, function(identificationCharacter, index) {
        identificationNumberInNotation[notationCharacters[index]] = identificationCharacter;
    });

    return identificationNumberInNotation;
}

/**
 * Perform modulus check
 * @param  {Object}   identificationNumberInNotation   The identification number in modulus checking notation
 * @param  {Object}   checkRecord                      The weight record
 * @param  {Object}   modulusCheckSubstitution         The current check substitution record
 * @param  {Number}   modulusBy                        The number to modulus by
 * @param  {Object}   [previousCheckRecord]            The previous check record (if relevant)
 * @param  {Boolean}  [valid]                          True if valid in previous checks
 *
 * @return  {Boolean}  True if check passed
 */
function modulusCheck(identificationNumberInNotation, checkRecord, modulusCheckSubstitution, modulusBy, previousCheckRecord, valid) {
    const exception = checkRecord.exception;
    let sum = 0,
        desiredRemainder = 0,
        checkDigit;

    if (exception) {
        const exceptionResult = performException(identificationNumberInNotation, checkRecord, modulusCheckSubstitution, previousCheckRecord, valid);

        if (_.isBoolean(exceptionResult)) {
            return exceptionResult;
        } else if (_.isObject(exceptionResult)) {
            sum = exceptionResult.sum;
            desiredRemainder = exceptionResult.desiredRemainder;
            identificationNumberInNotation = exceptionResult.identificationNumberInNotation;
            checkRecord = exceptionResult.checkRecord;
        }
    }
    _.each(identificationNumberInNotation, function(identificationCharacter, key) {
        const characterInNotation = key + '_multiplier';

        sum += identificationCharacter * checkRecord[characterInNotation];
    });

    return isModulusCheckPass(identificationNumberInNotation, sum, modulusBy, desiredRemainder, checkRecord);
}

/**
 * Perform the given exception
 * @param  {Object}   identificationNumberInNotation   The identification number in modulus checking notation
 * @param  {Object}   checkRecord                      The current check record
 * @param  {Object}   modulusCheckSubstitution         The current check substitution record
 * @param  {Object}   [previousCheckRecord]            The previous check record (if relevant)
 * @param  {Boolean}  [valid]                          True if valid in previous checks
 *
 * @return  {Boolean|Object}  Updated sum or true if exception causes validation pass
 */
function performException(identificationNumberInNotation, checkRecord, modulusCheckSubstitution, previousCheckRecord, valid) {
    const hasPreviousCheckRecord = !!previousCheckRecord;
    let sum = 0,
        desiredRemainder = 0;

    switch (checkRecord.exception) {
        case 1:
            sum += 27;
            break;
        case 2:
            checkRecord = _.merge(checkRecord, getException2Weights());
            break;
        case 3:
            if (_.includes([6, 9], identificationNumberInNotation['c'])) {
                return true;
            }
            break;
        case 4:
            if (identificationNumberInNotation['a'] !== 0 && identificationNumberInNotation['g'] !== 9) {
                desiredRemainder = parseInt(identificationNumberInNotation['g'].toString() + identificationNumberInNotation['h'].toString(), 10);
            }
            break;
        case 5:
            if (modulusCheckSubstitution) {
                identificationNumberInNotation = substituteSortCode(identificationNumberInNotation, modulusCheckSubstitution.substitute);
            }
            break;
        case 6:
            if (_.includes([4, 5, 6, 7, 8], identificationNumberInNotation['a']) && identificationNumberInNotation['g'] === identificationNumberInNotation['h']) {
                return true;
            }
            break;
        case 7:
            if (identificationNumberInNotation['g'] === 9) {
                identificationNumberInNotation = zeroiseUToB(identificationNumberInNotation);
            }
            break;
        case 8:
            identificationNumberInNotation = substituteSortCode(identificationNumberInNotation, 090126);
            break;
        case 9:
            if (hasPreviousCheckRecord && previousCheckRecord.exception === 2) {
                if (valid) {
                    return true;
                }
                if (identificationNumberInNotation['a'] !== 0 && identificationNumberInNotation['g'] !== 9) {
                    identificationNumberInNotation = substituteSortCode(identificationNumberInNotation, 309634);
                    checkRecord = _.merge(checkRecord, getException9Weights());
                }
            }
            break;
        case 10:
            const ab = identificationNumberInNotation['a'].toString() + identificationNumberInNotation['b'].toString();
            if ((ab === '09' || ab == '99') && identificationNumberInNotation['g'] === 9) {
                identificationNumberInNotation = zeroiseUToB(identificationNumberInNotation);
            }
            break;
        case 11:
            if (hasPreviousCheckRecord && previousCheckRecord.exception === 10 && valid) {
                return true;
            }
            break;
        case 12:
        case 13:
            // do nothing
            break;
        case 14:
            if (hasPreviousCheckRecord) {
                if (!_.includes([0, 1, 9], identificationNumberInNotation['h'])) {
                    return false;
                } else {
                    identificationNumberInNotation = removeLastDigit(identificationNumberInNotation);
                }
            }
            break;
    }

    return {
        identificationNumberInNotation: identificationNumberInNotation,
        sum: sum,
        desiredRemainder: desiredRemainder,
        checkRecord: checkRecord
    };
}

/**
 * Substitute the sort code of the given id number in notation
 * @param  {Object}   identificationNumberInNotation   The identification number in modulus checking notation
 * @param  {Number}   sortCodeSubstitute               The sort code substitute
 *
 * @return  {Object}  The updated id number in notation
 */
function substituteSortCode(identificationNumberInNotation, sortCodeSubstitute) {
    const positionsToSubstitute = ['u', 'v', 'w', 'x', 'y', 'z'];

    _.each(positionsToSubstitute, function(position, index) {
        identificationNumberInNotation[position] = sortCodeSubstitute.toString().charAt(index);
    });

    return identificationNumberInNotation;
}

/**
 * Zeroise u-b for the given id number
 * @param  {Object}   identificationNumberInNotation   The identification number in modulus checking notation
 *
 * @return  {Object}  The updated id number with u-b zeroised
 */
function zeroiseUToB(identificationNumberInNotation) {
    const positionsToZeroise = ['u', 'v', 'w', 'x', 'y', 'z', 'a', 'b'];

    _.each(positionsToZeroise, function(position) {
        identificationNumberInNotation[position] = 0;
    });

    return identificationNumberInNotation;
}

/**
 * Remove the last digit and append a 0 to the start of the account number
 * @param  {Object}   identificationNumberInNotation   The identification number in modulus checking notation
 *
 * @return  {Object}  The updated id number with last digit removed and 0 appended to start
 */
function removeLastDigit(identificationNumberInNotation) {
    const clonedIdentificationNumberInNotation = _.clone(identificationNumberInNotation),
        charactersToChange = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'];

    _.each(charactersToChange, function(character) {
        if (character === 'a') {
            clonedIdentificationNumberInNotation[character] = 0;
        } else {
            const currentCharacterIndex = _.indexOf(charactersToChange, character);

            clonedIdentificationNumberInNotation[character] = identificationNumberInNotation[charactersToChange[currentCharacterIndex - 1]];
        }
    });

    return clonedIdentificationNumberInNotation;
}

/**
 * Get the weights (multipliers) for exception 2
 *
 * @return  {Object}  The weights (multipliers) for exception 2
 */
function getException2Weights() {
    return {
        u_multiplier: 0,
        v_multiplier: 0,
        w_multiplier: 1,
        x_multiplier: 2,
        y_multiplier: 5,
        z_multiplier: 3,
        a_multiplier: 6,
        b_multiplier: 4,
        c_multiplier: 8,
        d_multiplier: 7,
        e_multiplier: 10,
        f_multiplier: 9,
        g_multiplier: 3,
        h_multiplier: 1
    };
}

/**
 * Get the weights (multipliers) for exception 9
 *
 * @return  {Object}  The weights (multipliers) for exception 9
 */
function getException9Weights() {
    return {
        u_multiplier: 0,
        v_multiplier: 0,
        w_multiplier: 0,
        x_multiplier: 0,
        y_multiplier: 0,
        z_multiplier: 0,
        a_multiplier: 0,
        b_multiplier: 0,
        c_multiplier: 8,
        d_multiplier: 7,
        e_multiplier: 10,
        f_multiplier: 9,
        g_multiplier: 3,
        h_multiplier: 1
    };
}

/**
 * Is the modulus check a pass based on the calculated results?
 * @param  {Object}   identificationNumberInNotation   The identification number in modulus checking notation
 * @param  {Number}   sum                        The sum of the identification numbers with multipliers applied
 * @param  {Number}   modulusBy                  The number to modulus by
 * @param  {Number}   desiredRemainder           The desired remainder
 * @param  {Object}   checkRecord                The current check record
 *
 * @return  {Object}  The updated id number in notation
 */
function isModulusCheckPass(identificationNumberInNotation, sum, modulusBy, desiredRemainder, checkRecord) {
    const remainder = sum % modulusBy;

    switch (checkRecord.exception) {
        case 5:
            if (checkRecord.mod_check === 'DblAl') {
                if (remainder === 0 && identificationNumberInNotation['h'] === 0) {
                    return true;
                }
                return (10 - remainder) === identificationNumberInNotation['h'];
            } else {
                if (remainder === 0 && identificationNumberInNotation['g'] === 0) {
                    return true;
                }
                return (11 - remainder) === identificationNumberInNotation['g'];
            }
        default:
            return remainder === desiredRemainder && sum !== 0;
    }
}

/**
 * Replace all occurences within a string with another string value
 *
 * @param {String} string The string to replace the characters in
 * @param {String} replace The value to replace in the string
 * @param {String} replaceWith The value to use as the replacement
 *
 * @return {String} The updated string with the values replaced
 */
function replaceAll(string, replace, replaceWith) {
    return string.split(replace).join(replaceWith);
}

/**
 * Returns an array of int or float from an array of number strings
 * @param {String[]} array   The array of strings
 * @param {Boolean}  [isFloat] True if array of floats
 *
 * @return {Number[]} The array of parsed integers or float
 */
function parseNumberArray(array, isFloat) {
    return _.map(array, function(arrayItem) {
        return isFloat ? parseFloat(arrayItem) : parseInt(arrayItem, 10);
    });
}